@@ -56,7 +56,7 @@ enum {
 
 /**
  * SoupBuffer:
- * @data: the data
+ * @data: (type gpointer): the data
  * @length: length of @data
  *
  * A data buffer, generally used to represent a chunk of a
@@ -108,6 +108,28 @@ soup_buffer_new (SoupMemoryUse use, gconstpointer data, gsize length)
 	return (SoupBuffer *)priv;
 }
 
+/**
+ * soup_buffer_new_take:
+ * @data: (array length=length) (transfer full): data
+ * @length: length of @data
+ *
+ * Creates a new #SoupBuffer containing @length bytes from @data.
+ *
+ * This function is exactly equivalent to soup_buffer_new() with
+ * %SOUP_MEMORY_TAKE as first argument; it exists mainly for
+ * convenience and simplifying language bindings.
+ *
+ * Return value: the new #SoupBuffer.
+ *
+ * Since: 2.32
+ * Rename to: soup_buffer_new
+ **/
+SoupBuffer *
+soup_buffer_new_take (guchar *data, gsize length)
+{
+	return soup_buffer_new (SOUP_MEMORY_TAKE, data, length);
+}
+
 /**
  * soup_buffer_new_subbuffer:
  * @parent: the parent #SoupBuffer
@@ -193,17 +215,39 @@ soup_buffer_new_with_owner (gconstpointer  data, gsize length,
  * Gets the "owner" object for a buffer created with
  * soup_buffer_new_with_owner().
  *
- * Return value: the owner pointer
+ * Return value: (transfer none): the owner pointer
  **/
 gpointer
 soup_buffer_get_owner (SoupBuffer *buffer)
 {
 	SoupBufferPrivate *priv = (SoupBufferPrivate *)buffer;
 
-	g_return_val_if_fail (priv->use == SOUP_MEMORY_OWNED, NULL);
+	g_return_val_if_fail ((int)priv->use == (int)SOUP_MEMORY_OWNED, NULL);
 	return priv->owner;
 }
 
+/**
+ * soup_buffer_get_data:
+ * @buffer: a #SoupBuffer
+ * @data: (out) (array length=length) (transfer none): the pointer
+ * to the buffer data is stored here
+ * @length: (out): the length of the buffer data is stored here
+ *
+ * This function exists for use by language bindings, because it's not
+ * currently possible to get the right effect by annotating the fields
+ * of #SoupBuffer.
+ *
+ * Since: 2.32
+ */
+void
+soup_buffer_get_data (SoupBuffer     *buffer,
+		      const guint8  **data,
+		      gsize          *length)
+{
+	*data = (const guint8 *)buffer->data;
+	*length = buffer->length;
+}
+
 /**
  * soup_buffer_copy:
  * @buffer: a #SoupBuffer
@@ -334,25 +378,33 @@ soup_message_body_new (void)
  * @body: a #SoupMessageBody
  * @accumulate: whether or not to accumulate body chunks in @body
  *
- * Sets or clears the accumulate flag on @body. (The default value
- * is %TRUE.)
+ * Sets or clears the accumulate flag on @body. (The default value is
+ * %TRUE.) If set to %FALSE, @body's %data field will not be filled in
+ * after the body is fully sent/received, and the chunks that make up
+ * @body may be discarded when they are no longer needed.
  *
- * If you set this flag to %FALSE on an "incoming" message body (that
- * is, the %response_body of a client-side message, or %request_body
- * of a server-side message), this will cause each chunk of the body
- * to be discarded after its corresponding #SoupMessage::got_chunk
- * signal is emitted. (This is equivalent to setting the deprecated
- * %SOUP_MESSAGE_OVERWRITE_CHUNKS flag on the message.)
+ * In particular, if you set this flag to %FALSE on an "incoming"
+ * message body (that is, the %response_body of a client-side message,
+ * or %request_body of a server-side message), this will cause each
+ * chunk of the body to be discarded after its corresponding
+ * #SoupMessage::got_chunk signal is emitted. (This is equivalent to
+ * setting the deprecated %SOUP_MESSAGE_OVERWRITE_CHUNKS flag on the
+ * message.)
  *
- * If you set this flag to %FALSE on an "outgoing" message body (the
- * %request_body of a client-side message, or %response_body of a
- * server-side message), it will cause each chunk of the body to be
+ * If you set this flag to %FALSE on the %response_body of a
+ * server-side message, it will cause each chunk of the body to be
  * discarded after its corresponding #SoupMessage::wrote_chunk signal
  * is emitted.
  *
- * In either case, @body's %data field will not be filled in after the
- * body is fully sent/received, since the body data will no longer be
- * available
+ * (If you set the flag to %FALSE on the %request_body of a
+ * client-side message, it will block the accumulation of chunks into
+ * @body's %data field, but it will not cause the chunks to be
+ * discarded after being written like in the server-side
+ * %response_body case, because the request body needs to be kept
+ * around in case the request needs to be sent a second time due to
+ * redirection or authentication.)
+ *
+ * Since: 2.4.1
  **/
 void
 soup_message_body_set_accumulate (SoupMessageBody *body,
@@ -371,6 +423,8 @@ soup_message_body_set_accumulate (SoupMessageBody *body,
  * soup_message_body_set_accumulate() for details.
  *
  * Return value: the accumulate flag for @body.
+ *
+ * Since: 2.4.1
  **/
 gboolean
 soup_message_body_get_accumulate (SoupMessageBody *body)
@@ -403,7 +457,7 @@ append_buffer (SoupMessageBody *body, SoupBuffer *buffer)
  * soup_message_body_append:
  * @body: a #SoupMessageBody
  * @use: how to use @data
- * @data: data to append
+ * @data: (array length=length) (element-type guint8): data to append
  * @length: length of @data
  *
  * Appends @length bytes from @data to @body according to @use.
@@ -418,6 +472,28 @@ soup_message_body_append (SoupMessageBody *body, SoupMemoryUse use,
 		g_free ((gpointer)data);
 }
 
+/**
+ * soup_message_body_append_take:
+ * @body: a #SoupMessageBody
+ * @data: (array length=length) (transfer full): data to append
+ * @length: length of @data
+ *
+ * Appends @length bytes from @data to @body.
+ *
+ * This function is exactly equivalent to soup_message_body_apppend()
+ * with %SOUP_MEMORY_TAKE as second argument; it exists mainly for
+ * convenience and simplifying language bindings.
+ *
+ * Since: 2.32
+ * Rename to: soup_message_body_append
+ **/
+void
+soup_message_body_append_take (SoupMessageBody *body,
+			       guchar *data, gsize length)
+{
+	soup_message_body_append(body, SOUP_MEMORY_TAKE, data, length);
+}
+
 /**
  * soup_message_body_append_buffer:
  * @body: a #SoupMessageBody
@@ -578,6 +654,8 @@ soup_message_body_get_chunk (SoupMessageBody *body, goffset offset)
  *
  * This is a low-level method which you should not normally need to
  * use.
+ *
+ * Since: 2.4.1
  **/
 void
 soup_message_body_got_chunk (SoupMessageBody *body, SoupBuffer *chunk)
@@ -593,15 +671,18 @@ soup_message_body_got_chunk (SoupMessageBody *body, SoupBuffer *chunk)
 /**
  * soup_message_body_wrote_chunk:
  * @body: a #SoupMessageBody
- * @chunk: a #SoupBuffer received from the network
+ * @chunk: a #SoupBuffer returned from soup_message_body_get_chunk()
  *
  * Handles the #SoupMessageBody part of writing a chunk of data to the
  * network. Normally this is a no-op, but if you have set @body's
- * accumulate flag to %FALSE, then this will cause @chunk (and any
- * chunks preceding it in @body) to be discarded to free up memory.
+ * accumulate flag to %FALSE, then this will cause @chunk to be
+ * discarded to free up memory.
  *
- * This is a low-level method which you should not normally need to
- * use.
+ * This is a low-level method which you should not need to use, and
+ * there are further restrictions on its proper use which are not
+ * documented here.
+ *
+ * Since: 2.4.1
  **/
 void
 soup_message_body_wrote_chunk (SoupMessageBody *body, SoupBuffer *chunk)
@@ -612,15 +693,16 @@ soup_message_body_wrote_chunk (SoupMessageBody *body, SoupBuffer *chunk)
 	if (priv->accumulate)
 		return;
 
-	do {
-		chunk2 = priv->chunks->data;
-		priv->chunks = g_slist_remove (priv->chunks, chunk2);
-		priv->base_offset += chunk2->length;
-		soup_buffer_free (chunk2);
-	} while (priv->chunks && chunk2 != chunk);
+	chunk2 = priv->chunks->data;
+	g_return_if_fail (chunk->length == chunk2->length);
+	g_return_if_fail (chunk == chunk2 || ((SoupBufferPrivate *)chunk2)->use == SOUP_MEMORY_TEMPORARY);
 
+	priv->chunks = g_slist_remove (priv->chunks, chunk2);
 	if (!priv->chunks)
 		priv->last = NULL;
+
+	priv->base_offset += chunk2->length;
+	soup_buffer_free (chunk2);
 }
 
 static SoupMessageBody *
