@@ -98,19 +98,27 @@ soup_multipart_new (const char *mime_type)
 }
 
 static const char *
-find_boundary (const char *start, const char *boundary, int boundary_len)
+find_boundary (const char *start, const char *end,
+	       const char *boundary, int boundary_len)
 {
-	const char *b, *end;
+	const char *b;
 
-	end = start + 2;
-	while ((b = strstr (end, boundary))) {
-		end = b + boundary_len;
-		if (b[-1] == '-' && b[-2] == '-' &&
-		    (b == start + 2 || (b[-3] == '\n' && b[-4] == '\r'))) {
-			if ((end[0] == '-' && end[1] == '-') ||
-			    (end[0] == '\r' && end[1] == '\n'))
-				return b - 2;
-		}
+	for (b = memchr (start, '-', end - start);
+	     b && b + boundary_len + 4 < end;
+	     b = memchr (b + 2, '-', end - (b + 2))) {
+		/* Check for "--boundary" */
+		if (b[1] != '-' ||
+		    memcmp (b + 2, boundary, boundary_len) != 0)
+			continue;
+
+		/* Check that it's at start of line */
+		if (!(b == start || (b[-1] == '\n' && b[-2] == '\r')))
+			continue;
+
+		/* Check for "--" or "\r\n" after boundary */
+		if ((b[boundary_len + 2] == '-' && b[boundary_len + 3] == '-') ||
+		    (b[boundary_len + 2] == '\r' && b[boundary_len + 3] == '\n'))
+			return b;
 	}
 	return NULL;
 }
@@ -136,7 +144,7 @@ soup_multipart_new_from_message (SoupMessageHeaders *headers,
 	GHashTable *params;
 	int boundary_len;
 	SoupBuffer *flattened;
-	const char *start, *split, *end;
+	const char *start, *split, *end, *body_end;
 	SoupMessageHeaders *part_headers;
 	SoupBuffer *part_body;
 
@@ -155,11 +163,13 @@ soup_multipart_new_from_message (SoupMessageHeaders *headers,
 	g_hash_table_destroy (params);
 
 	flattened = soup_message_body_flatten (body);
+	body_end = flattened->data + flattened->length;
 	boundary = multipart->boundary;
 	boundary_len = strlen (boundary);
 
 	/* skip preamble */
-	start = find_boundary (flattened->data, boundary, boundary_len);
+	start = find_boundary (flattened->data, body_end,
+			       boundary, boundary_len);
 	if (!start) {
 		soup_multipart_free (multipart);
 		soup_buffer_free (flattened);
@@ -167,7 +177,8 @@ soup_multipart_new_from_message (SoupMessageHeaders *headers,
 	}
 
 	while (start[2 + boundary_len] != '-') {
-		end = find_boundary (start + 2 + boundary_len, boundary, boundary_len);
+		end = find_boundary (start + 2 + boundary_len, body_end,
+				     boundary, boundary_len);
 		if (!end) {
 			soup_multipart_free (multipart);
 			soup_buffer_free (flattened);
@@ -374,20 +385,6 @@ soup_multipart_append_form_file (SoupMultipart *multipart,
 					     content_type);
 	}
 
-	/* The HTML spec says we need to set Content-Transfer-Encoding
-	 * if the data is not 7bit. It probably doesn't actually
-	 * matter...
-	 */
-	if (content_type && strncmp (content_type, "text/", 5) != 0) {
-		soup_message_headers_append (headers,
-					     "Content-Transfer-Encoding",
-					     "binary");
-	} else {
-		soup_message_headers_append (headers,
-					     "Content-Transfer-Encoding",
-					     "8bit");
-	}
-
 	g_ptr_array_add (multipart->headers, headers);
 	g_ptr_array_add (multipart->bodies, soup_buffer_copy (body));
 }
@@ -412,21 +409,22 @@ soup_multipart_to_message (SoupMultipart *multipart,
 	SoupMessageHeadersIter iter;
 	const char *name, *value;
 	GString *str;
-	char *content_type;
+	GHashTable *params;
 	int i;
 
-	content_type = g_strdup_printf ("%s; boundary=\"%s\"",
-					multipart->mime_type,
-					multipart->boundary);
-	soup_message_headers_replace (dest_headers, "Content-Type",
-				      content_type);
-	g_free (content_type);
+	params = g_hash_table_new (g_str_hash, g_str_equal);
+	g_hash_table_insert (params, "boundary", multipart->boundary);
+	soup_message_headers_set_content_type (dest_headers,
+					       multipart->mime_type,
+					       params);
+	g_hash_table_destroy (params);
 
 	for (i = 0; i < multipart->bodies->len; i++) {
 		part_headers = multipart->headers->pdata[i];
 		part_body = multipart->bodies->pdata[i];
 
-		str = g_string_new ("\r\n--");
+		str = g_string_new (i == 0 ? NULL : "\r\n");
+		g_string_append (str, "--");
 		g_string_append (str, multipart->boundary);
 		g_string_append (str, "\r\n");
 		soup_message_headers_iter_init (&iter, part_headers);
